/* STLC2590 radio driver for Linux (C) 2008 Lucas De Marchi, Cristina Chang (lucas.de.marchi@gmail.com, titinachang@gmail.com)
 *
 * This driver uses V4L2 API and is coded according to the specs given by ST.
 *
 * Besides the protocol changes, this is (loosely) based on:
 *			RadioTrack II driver for Linux radio support (C) 1998 Ben Pfaff
 *			GemTek radio card driver for Linux (C) 1998 Jonas Munsin
 *			RadioTrack I/RadioReveal (C) 1997 M. Kirkwood
 *			ADS Cadet AM/FM Radio Card
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 */

#include	<linux/init.h>			/* Initdata		*/
#include	<linux/module.h>		/* Modules 		*/

#include  	<linux/delay.h>

#include	<asm/uaccess.h>   		/* copy to/from user */

#include	<linux/videodev2.h>		/* kernel radio structs */
#include	<media/v4l2-common.h>		/* v4l2 functions */

#include	<linux/version.h>   		/* for KERNEL_VERSION macro*/




/* --------   Nomadik headers ---------  */

#include	<asm/arch/nhk15_devices.h>
#include	<asm/arch/audiocodec.h>
#include	<asm/arch/i2c.h>
#include  	<asm/arch/pexp.h>
#include	<asm/arch/gpio.h>

#include 	<asm/arch/debug.h>

#define RADIO_VERSION KERNEL_VERSION(0,0,12)

#ifndef STLC2590_DEBUG
#define STLC2590_DEBUG 0
#else
#warning "DEBUG ENABLED"
#endif

#define NMDK_STLC2590_NAME "FM"
#define NMDK_DEBUG STLC2590_DEBUG
#define NMDK_DEBUG_PFX NMDK_STLC2590_NAME
#define NMDK_DBG KERN_ERR


static wait_queue_head_t read_queue;


/*Module parameters*/

 /*
  * 0 - 200 kHz (USA, Australia)
  * 1 - 100 kHz (Europe, Japan)		[DEFAULT]
  * 2 - 50  kHz ( ?? )
  */
#define STLC2590_DEFAULT_CHANNEL_SPACING 0x01 // 100 kHz - Europe
unsigned int channel_spacing = STLC2590_DEFAULT_CHANNEL_SPACING;
module_param(channel_spacing, uint, S_IRUGO);
MODULE_PARM_DESC(channel_spacing, "The space between each radio station, according to which region the user is in");

 /*
  * 00 = 87.5 - 108 MHz (US /EUROPE) 	[DEFAULT]
  * 01 = 76   - 108 MHz (Japan wide band)
  * 10 = 76   - 90  MHz (Japan)
  */
#define STLC2590_DEFAULT_BAND_LIMITS 0x00 // 87.5 - 108 MHz - Europe
unsigned int band_limits = STLC2590_DEFAULT_BAND_LIMITS;
module_param(band_limits, uint, S_IRUGO);
MODULE_PARM_DESC(band_limits, "It defines the upper and lower band limits, according to which region the user is in");

#define STLC2590_JAPAN_WIDE_BAND_LIMITS 0x01
#define STLC2590_JAPAN_BAND_LIMITS 0x02

#define STLC2590_DEFAULT_RDS_ENABLE 0x01
unsigned int rds_enable = STLC2590_DEFAULT_RDS_ENABLE;
module_param(rds_enable, uint, S_IRUGO);
MODULE_PARM_DESC(rds_enable, "Enable or disable rds functions");


#define STLC2590_DEFAULT_STEREO_BLEND 0x00
unsigned int stereo_blend = STLC2590_DEFAULT_STEREO_BLEND;
module_param(stereo_blend, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(stereo_blend,  "Lower bound of stereo blend");


#define  STLC2590_DEFAULT_EXTERNAL_AMP_ENABLE 0x00;
unsigned int external_amp_enable = STLC2590_DEFAULT_EXTERNAL_AMP_ENABLE;
module_param(external_amp_enable, uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(external_amp_enable,  "Enable external amplifiers");

#define RDS_REG_LEN 3
#define RDS_BUFFER 4
static unsigned char rdsbuf[RDS_BUFFER*RDS_REG_LEN];
static int rdsp = 0;

char pname[9]="        \0";
unsigned int rdstest = 0;
/*
 * As these parameters are configured via V4L2, there's no need to put them as module parameters
 */
#define STLC2590_DEFAULT_VOLUME 0xFF
#define STLC2590_DEFAULT_MUTE 0x01


#define STLC2590_CONV_SIGNALSTRENGTH 873
#define STLC2590_CONV_VOLUME 4369
#define STLC2590_CONV_FREQUENCY 160

typedef enum {
	STLC2590_DEVICEID_UP=0x00,	STLC2590_DEVICEID_LOW=0x01,
	STLC2590_CHIPID_UP=0x02,	STLC2590_CHIPID_LOW=0x03,
	STLC2590_POWERCFG_UP=0x04,	STLC2590_POWERCFG_LOW=0x05,
	STLC2590_CHANNEL_UP=0x06,	STLC2590_CHANNEL_LOW=0x07,
	STLC2590_SYSCONFIG1_UP=0x08,	STLC2590_SYSCONFIG1_LOW=0x09,
	STLC2590_SYSCONFIG2_UP=0x0A,	STLC2590_SYSCONFIG2_LOW=0x0B,
	STLC2590_SYSCONFIG3_UP=0x0C,	STLC2590_SYSCONFIG3_LOW=0x0D,
	STLC2590_TEST1_UP=0x0E,		STLC2590_TEST1_LOW=0x0F,
	STLC2590_TEST2_UP=0x10,		STLC2590_TEST2_LOW=0x11,
	STLC2590_BOOTCONFIG_UP=0x12,	STLC2590_BOOTCONFIG_LOW=0x13,
	STLC2590_STATUSRSSI_UP=0x14,	STLC2590_STATUSRSSI_LOW=0x15,
	STLC2590_READCHAN_UP=0x16,	STLC2590_READCHAN_LOW=0x17,
	STLC2590_RDSA_UP=0x18,		STLC2590_RDSA_LOW=0x19,
	STLC2590_RDSB_UP=0x1A,		STLC2590_RDSB_LOW=0x1B,
	STLC2590_RDSC_UP=0x1C,		STLC2590_RDSC_LOW=0x1D,
	STLC2590_RDSD_UP=0x1E,		STLC2590_RDSD_LOW=0x1F,

} t_stlc2590_regs;

typedef enum {
	STLC2590_PIN_MUTE=6,
	STLC2590_PIN_SMUTE=7,
	STLC2590_PIN_CHANNEL_SPACING=4,
	STLC2590_PIN_BAND_LIMIT=6,
	STLC2590_PIN_TUNE=7,
	STLC2590_PIN_STC=6,
	STLC2590_PIN_SEEK=0,
	STLC2590_PIN_POWERCFG_0=4,
	STLC2590_PIN_RDSENABLE=4,
	STLC2590_PIN_GPIO3=4,
	STLC2590_PIN_BLNDADJ=6,
	STLC2590_PIN_RDSINTENABLE=7,
	STLC2590_PIN_RDSM=3
} t_stlc2590_regs_pin;


static int radio_nr = -1; /*Let V4L2 assign the minor number*/


struct t_radio_device {
	uint8 regs[32];
};



//static unsigned long mixchann = 0;
//module_param(mixchann, ulong, 0);
//MODULE_PARM_DESC(mixchann, "Mixer channel");

static struct v4l2_queryctrl radio_qctrl[] = {
	{
		.id 					= V4L2_CID_AUDIO_MUTE,
		.name 					= "Mute",
		.minimum				= 0,
		.maximum 				= 1,
		.default_value 	= 1,
		.type					= V4L2_CTRL_TYPE_BOOLEAN,
	},{
		.id 					= V4L2_CID_AUDIO_VOLUME,
		.name 					= "Volume",
		.minimum				= 0,
		.maximum 				= 15,
		.step					= 1,
		.default_value 				= 15,
		.type					= V4L2_CTRL_TYPE_INTEGER,
	}
};


void printregs(uint8 * regs)
{
	int i, f;
	for(i = 0; i < 32; i+=2)										
		nmdk_info("Register [0x%02x]: 0x%02X 0x%02x", (int)(i/2),regs[i],regs[i+1] );

	f =  (((regs[22] & 0x03) << 8 ) |  regs[23]) + 875;
	nmdk_info("Readchannel: %d", f);
	nmdk_info("Tune bit: %s", (regs[6] & 0x80) ? "1":"0");
	nmdk_info("STC bit:  %s",  (regs[20] & 0x40) ? "1":"0");
	nmdk_info("SF/BL bit: %s", (regs[20] & 0x20) ? "1":"0");
	return;
}


/*
 * This is a workaround to bypass alsa configuration. Remove as soon as alsa driver for NHK15 is fixed.
 * It configures all registers on audiocodec STW5095 to route sound.
 */
static int init_audio_codec(void)
{
#ifdef CONFIG_NOMADIK_NHK15
	uint8 reg;
	int i2c_error=0;

	// software reset
	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 30, 1);
	reg |= 0x80;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 30, 1); 

	msleep(100);

	// AMCKSIN = 0
	// CKRANGE = 100
	reg = 0x04; 
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 30, 1); 
		
	// Analog Functions
	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 17, 1);
	reg |= 0x04;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 17, 1); 

	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 18, 1);
	reg = 0x18;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 18, 1); 

	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 19, 1);
	reg = 0x40;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 19, 1); 

	// Power
	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 1, 1);
	reg |= 0x03;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 1, 1); 
	
	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 2, 1);
	reg |= 0xFB;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 2, 1); 
	
	i2c_error += nomadik_i2c_read_register( I2C_AUDIO_CODEC_CLIENT, &reg, 0, 1);
	reg |= 0xE0;
	i2c_error += nomadik_i2c_write_register(I2C_AUDIO_CODEC_CLIENT, &reg, 0, 1); 
	
	if(i2c_error != 0){
		nmdk_error("ERROR while configuring routing on audiocodec");
		return -1;
	}


	// External amplifiers STBYn
	i2c_error += STMPE2401_SetGpioDir(0,EGPIO_PIN_12,STMPE2401_GPIO_OUT);
	i2c_error += STMPE2401_SetGpioVal(0,EGPIO_PIN_12,external_amp_enable?1:0);

	if(i2c_error != 0){
		nmdk_error("ERROR while configuring port expander pins");
		return -1;
	}
#endif	
	return 0;
}

/*
 * This function must be called every time a configuration is changed. It calls I2C write function to transfer the
 * configuration values to chip.
 * @param dev is used to take the regs vector. Before call this function, update desired register values.
 */
static int stlc2590_refresh_write(struct t_radio_device *dev)
{
	uint8 * regs = dev->regs;
	if(nomadik_i2c_write_register(I2C_FM_TUNER_CLIENT, &(regs[4]), 0,10))
		return -1;
	return 0;
}

/*
 * This function reads all registers from chip and update the corresponing values on dev->regs array.
 */
static int stlc2590_refresh_read (struct t_radio_device *dev)
{
	uint8 data[32];
	int err, i;

	memset(data, 0, 32);
	err = nomadik_i2c_read_register(I2C_FM_TUNER_CLIENT, data, 0, 32);
	if(err)
		return -1;
	
	for (i = 0; i<32; i++)
		dev->regs[(i + 20) % 32]=data[i];
	
	return 0;
}


/*
 * Enable or disable rds functions
 * @param enable:
 * 	0 - RDS disabled
 * 	1 - RDS enabled
 * @param flush:
 * 	0 - Only vars are changed
 * 	1 - Vars are changed and written to chip
 */
static int stlc2590_set_rdsenable(struct t_radio_device *dev, int enable, int flush)
{

	if(enable){
		dev->regs[STLC2590_SYSCONFIG1_UP] |= (uint8)(1 << STLC2590_PIN_RDSENABLE);
		dev->regs[STLC2590_SYSCONFIG1_UP] |= (uint8)(1 << STLC2590_PIN_RDSINTENABLE);
		dev->regs[STLC2590_POWERCFG_UP] |= (uint8) (1 << STLC2590_PIN_RDSM);
	} else {
		dev->regs[STLC2590_SYSCONFIG1_UP] &= ~((uint8)(1 << STLC2590_PIN_RDSENABLE));
		dev->regs[STLC2590_SYSCONFIG1_UP] &= ~((uint8)(1 << STLC2590_PIN_RDSINTENABLE));
		dev->regs[STLC2590_POWERCFG_UP] &= ~((uint8) (1 << STLC2590_PIN_RDSM));
	}


	if(flush && stlc2590_refresh_write(dev)){
		nmdk_info("Error trying to change rds_enable bit");
		return -1;
	}
	return 0;
}

static int stlc2590_set_stereoindicator(struct t_radio_device *dev, int enable, int flush)
{

	dev->regs[STLC2590_SYSCONFIG1_LOW] &= (uint8) ~( 0x03 << STLC2590_PIN_GPIO3);
	if(enable)
		dev->regs[STLC2590_SYSCONFIG1_LOW] |= (uint8) ( 0x01 << STLC2590_PIN_GPIO3);

	if(flush && stlc2590_refresh_write(dev)){
		nmdk_info("Error trying to configure stereo indicator pins");
		return -1;
	}
	return 0;
}

static int stlc2590_get_stereoindicator(struct t_radio_device *dev)
{
	t_STMPE2401_error err;
	uint8 val;
//	stlc2590_refresh_read(dev);
	
	err = STMPE2401_GetGpioVal(0,EGPIO_PIN_14,&val);
	
	return (val);
}

static int stlc2590_set_stereoblend(struct t_radio_device *dev, uint8 val, int flush)
{
	if(val >=4) {
		nmdk_error ("Stereo blend value out of range");
		return -1;
	}

	val = ((val + 2) % 4); 
	nmdk_info("BLDADJ = %d;", val);

	val <<= STLC2590_PIN_BLNDADJ;
	
	dev->regs[STLC2590_SYSCONFIG1_LOW] &= ~(0x03 << STLC2590_PIN_BLNDADJ);
	dev->regs[STLC2590_SYSCONFIG1_LOW] |= val;

	if(flush && stlc2590_refresh_write(dev)){
		nmdk_info("Error trying to configure stereo blend level");
		return -1;
	}
	return 0;

}


/*
 * Mute or unmute the device
 * @param mute:
 * 	0 - Both Mute and Soft-Mute disabled
 * 	1 - Both Mute and Soft-Mute enabled
 * @param flush:
 * 	0 - Only vars are changed
 * 	1 - Vars are changed and written to chip
 */
static int stlc2590_set_mute (struct t_radio_device *dev, int mute, int flush)
{
	if(mute)
		dev->regs[STLC2590_POWERCFG_UP] &= ~((uint8)(1 << STLC2590_PIN_MUTE) | (uint8)(1<<STLC2590_PIN_SMUTE));
	else
		dev->regs[STLC2590_POWERCFG_UP] |= ((uint8)(1 << STLC2590_PIN_MUTE) | (uint8)(1<<STLC2590_PIN_SMUTE));
	
	if(flush && stlc2590_refresh_write(dev)){
		nmdk_info("Error trying to change mute bits");
		return -1;
	}
	return 0;
}

inline static int stlc2590_get_mute(struct t_radio_device *dev)
{
	stlc2590_refresh_read(dev);
	return ((dev->regs[STLC2590_POWERCFG_UP] & ((uint8)(1 << STLC2590_PIN_MUTE) | (uint8)(1<<STLC2590_PIN_SMUTE)))==0);
}


/*
 * Set the spacing between each channel
 * @param channel_spacing = {0,1,2}
 * 0 - 200 kHz (USA, Australia)
 * 1 - 100 kHz (Europe, Japan)
 * 2 - 50  kHz ( ?? )
 */
static int stlc2590_set_channel_spacing(struct t_radio_device *dev, uint8 channel_spacing, int flush)
{
	if(channel_spacing > 2)
		return -EINVAL;
	
	dev->regs[STLC2590_SYSCONFIG2_LOW] &= ~((uint8)(3 << STLC2590_PIN_CHANNEL_SPACING));  	 //XX00XXXX
	dev->regs[STLC2590_SYSCONFIG2_LOW] |= channel_spacing << STLC2590_PIN_CHANNEL_SPACING;

	if(flush && stlc2590_refresh_write(dev)){
		nmdk_error("Error trying to change channel spacing bits");
		return -1;
	}
	return 0;
}

/*
 * Returns channel spacing as MHz * 100   or 0 if error
 */
static uint16 stlc2590_get_channel_spacing(struct t_radio_device *dev)
{
	uint8 res;
	stlc2590_refresh_read(dev);
	res = dev->regs[STLC2590_SYSCONFIG2_LOW] >> STLC2590_PIN_CHANNEL_SPACING;
	res&=3;

	switch(res){
		case 0: //200 kHz (USA, Australia)
			return 20;
		case 1: //100 kHz (Europe, Japan)
			return 10;
		case 2: //50 kHz ( ?? )
			return 5;
		default:
			nmdk_error("Unknown channel spacing read from chip");
			return 0;
	}
}

/*
 * Set band limits for radio depending on region
 * @param band_limits = {0, 1, 2}
 * 00 = 87.5 - 108 MHz (US /EUROPE) --> CHIP Default
 * 01 = 76   - 108 MHz (Japan wide band)
 * 10 = 76   - 90  MHz (Japan)
 */
static int stlc2590_set_band_limits(struct t_radio_device *dev, uint8 band_limits, int flush)
{

	if(band_limits > 2)
		return -EINVAL;

	dev->regs[STLC2590_SYSCONFIG2_LOW] &= ~((uint8) 3 << STLC2590_PIN_BAND_LIMIT); //00XXXXXX
	dev->regs[STLC2590_SYSCONFIG2_LOW] |= band_limits<< STLC2590_PIN_BAND_LIMIT;

	if(flush && stlc2590_refresh_write(dev)){
		nmdk_error("Error trying to change band limit bits");
		return -1;
	}
	return 0;
}

/*
 * Returns band limits (in parameters rangelow and rangehigh) as MHz * 100
 */
static void stlc2590_get_band_limits(struct t_radio_device *dev, uint16 * rangelow, uint16 * rangehigh)
{
	uint8 res;
	stlc2590_refresh_read(dev);
	res = dev->regs[STLC2590_SYSCONFIG2_LOW] >> STLC2590_PIN_BAND_LIMIT;
	res &= 3;
	switch (res){
	case STLC2590_DEFAULT_BAND_LIMITS:
		*rangelow = 8750;
		*rangehigh = 10800;
		//v->rangelow = 1400000;   // 87.5 MHz / 62.5 Hz
		//v->rangehigh = 1728000;  // 108 Mhz / 62.5 Hz
		break;
	case STLC2590_JAPAN_WIDE_BAND_LIMITS:
		*rangelow = 7600;
		*rangehigh = 10800;
		//v->rangelow = 1216000;   // 76 MHz / 62.5 Hz
		//v->rangehigh = 1728000;  // 108 Mhz / 62.5 Hz
		break;
	case STLC2590_JAPAN_BAND_LIMITS:
		*rangelow = 7600;
		*rangehigh = 9000;
		//v->rangelow = 1400000;   // 76 MHz / 62.5 Hz
		//v->rangehigh = 1440000;  // 90 Mhz / 62.5 Hz
		break;
	}
	return;
}

/*
 * Set current radio volume
 * @param vol, 4 bit volume control
 */
static int stlc2590_set_volume(struct t_radio_device *dev, uint8 vol, int flush)
{

	vol &= ((uint8)0x0F); //map to 4bit control

	dev->regs[STLC2590_SYSCONFIG2_LOW] &= ~((uint8)0x0F);
	dev->regs[STLC2590_SYSCONFIG2_LOW] |= vol;

	if(flush && stlc2590_refresh_write(dev)){
		nmdk_error("Error trying to change band limit bits");
		return -1;
	}
	return 0;
}

/*
 * Get current radio volume
 */
inline static int stlc2590_get_volume(struct t_radio_device *dev)
{
	stlc2590_refresh_read(dev);
	//first 4 bits
	return (dev->regs[STLC2590_SYSCONFIG2_LOW] & 0x0F);
}



/*
 * Set current tuned frequency
 * Frequency as MHz * 100
 */
static int stlc2590_tune_frequency(struct t_radio_device *dev, uint16 freq)
{

	int i=0;
	uint8 * regs = dev->regs;
	uint16 nchann, rangelow, rangehigh;

	uint8 channelspc;
	stlc2590_refresh_read(dev);

	stlc2590_get_band_limits(dev, &rangelow, &rangehigh);
	channelspc = stlc2590_get_channel_spacing(dev);

	nchann = (freq - rangelow) / channelspc;


	
	stlc2590_set_stereoblend(dev, stereo_blend,0);
	
	nmdk_info("freq: %d, rangelow: %d; channelspc: %d; nchann: %d", freq, rangelow, channelspc, nchann);


	// Put SEEK bit down
	regs[STLC2590_POWERCFG_UP] &= ~((1<<STLC2590_PIN_SEEK) | (1 << STLC2590_PIN_POWERCFG_0));

	// Put RDS on verbose mode
	//regs[STLC2590_POWERCFG_UP] |= 0x08;
	
	//set TUNE bit low
	regs[STLC2590_CHANNEL_UP] &= (uint8) 0x03; // All bits, except that from channel goes to 0

	//write channel to tune
	regs[STLC2590_CHANNEL_UP] = (uint8) (nchann >> 8);
	regs[STLC2590_CHANNEL_LOW] = (uint8) (nchann & 0xFF);
	stlc2590_refresh_write(dev);
	msleep(100);
	stlc2590_refresh_read(dev);

	//set TUNE bit high
	regs[STLC2590_CHANNEL_UP] |= (uint8)(1<<STLC2590_PIN_TUNE);
	stlc2590_refresh_write(dev);


	//see whether device is tuned or not
	//Check if STC bit is set high
	i=0;
	do{
		//TODO: remove i
		msleep(100);
		stlc2590_refresh_read(dev);
		i++;
	} while(((regs[STLC2590_STATUSRSSI_UP] & (1<< STLC2590_PIN_STC))?0:1)  && (i < 20));


#if STLC2590_DEBUG
	nmdk_info("Iterations to tune: %d", i);
#endif


	//set TUNE bit down  - Complete tune operation
	regs[STLC2590_CHANNEL_UP] &= (uint8)0x03;
	stlc2590_refresh_write(dev);
	msleep(100);
	stlc2590_refresh_read(dev);
	

#if STLC2590_DEBUG
	printregs(dev->regs);
#endif

	if(i>=20){
		nmdk_error("Error trying to tune frequency");
		return -1;
	}


	return 0;
}

/*
 * Returns current frequency in MHz * 100
 * @param dev struct with device registers
 */
static uint16 stlc2590_get_frequency(struct t_radio_device *dev)
{
	uint8 channelspc;
	uint16 nchann, rangelow, rangehigh;

	stlc2590_refresh_read(dev);

	nchann = (uint16) ((dev->regs[STLC2590_READCHAN_UP] & 0x03) | dev->regs[STLC2590_READCHAN_LOW]);
	stlc2590_get_band_limits(dev, &rangelow, &rangehigh);
	channelspc = stlc2590_get_channel_spacing(dev);

	return (nchann * channelspc + rangelow);
}



/*
 * Returns the signal strength (rssi level)
 * According to hw specs this ranges from 0 dBuV to 75 dBuV (but it can return a value greater than this, without accuracy)
 */
static uint8 stlc2590_get_signalstrength(struct t_radio_device *dev)
{
	stlc2590_refresh_read(dev);
	return dev->regs[STLC2590_STATUSRSSI_LOW]>75?75:dev->regs[STLC2590_STATUSRSSI_LOW];
}




/* | | |
 * v v v  V4L2 stuff
 *
 */


/*
 * Reports functions and I/O supported methods by the device
 * DONE
 */
static int vidioc_querycap (struct file *file, void *priv,
		struct v4l2_capability *v)
{
	strlcpy(v->driver, "stlc2590", sizeof(v->driver));
	strlcpy(v->card, "STLC2590 Radio", sizeof(v->card));
	sprintf(v->bus_info, " ");
	v->version = RADIO_VERSION;
	v->capabilities = V4L2_CAP_TUNER | V4L2_CAP_READWRITE;
	return 0;
}

/*
 * Reports supported frequency range, signal status and other TUNER infos
 * Almost done
 */
static int vidioc_g_tuner (struct file *file, void *priv,
		struct v4l2_tuner *v)
{
	uint16 rangelow, rangehigh;
	struct video_device *dev = video_devdata(file);
	struct t_radio_device * radio_device = dev->priv;

	if (v->index > 0)
		return -EINVAL;


	strcpy(v->name, "FM");
	v->type = V4L2_TUNER_RADIO;
	
	stlc2590_get_band_limits(radio_device, &rangelow, &rangehigh);
	v->rangelow = rangelow * STLC2590_CONV_FREQUENCY;
	v->rangehigh = rangehigh * STLC2590_CONV_FREQUENCY;	
	
	v->rxsubchans =  stlc2590_get_stereoindicator(radio_device) ? V4L2_TUNER_SUB_STEREO : V4L2_TUNER_SUB_MONO;
	
	v->capability = V4L2_TUNER_CAP_LOW | V4L2_TUNER_CAP_STEREO;
	v->audmode = V4L2_TUNER_MODE_STEREO;
	
	// V4L2 signal strength ranges from 0 to 65535
	// Radio's ranges from 0 to 75.... converting:
	v->signal = stlc2590_get_signalstrength(radio_device) * STLC2590_CONV_SIGNALSTRENGTH;


	return 0;
}

/*
 * Index must no be set to any value other than 0
 * TODO: let user set "force mono" with v->audmode parameter
 */
static int vidioc_s_tuner (struct file *file, void *priv,
		struct v4l2_tuner *v)
{
	if(v->index > 0)
		return -EINVAL;
	return 0;
}

/*
 * Get current frequency
 * DONE
 */
static int vidioc_g_frequency (struct file *file, void *priv,
		struct v4l2_frequency *f)
{
	struct video_device *dev = video_devdata(file);
	struct t_radio_device * radio_device = dev->priv;
	
	uint16 freq, rangelow, rangehigh;

	freq = stlc2590_get_frequency(radio_device);
	stlc2590_get_band_limits(radio_device, &rangelow, &rangehigh);
	
	f->frequency = (freq + rangelow) * STLC2590_CONV_FREQUENCY; 	//convert in units of 62.5 Hz

	return 0;
}

/*
 * Set current frequency
 * DONE
 */
static int vidioc_s_frequency (struct file *file, void *priv,
		struct v4l2_frequency *f)
{
	uint16 freq;

	struct video_device *dev = video_devdata(file);
	struct t_radio_device * radio_device = dev->priv;

	freq = (uint16)(f->frequency / STLC2590_CONV_FREQUENCY);
	
	nmdk_info("Frequency: %d %d",f->frequency, freq);

	return(stlc2590_tune_frequency(radio_device, freq));
}

/*
 * Reports the available controls to applications (Volume, mute, etc)
 * DONE
 */
static int vidioc_queryctrl (struct file *file, void *priv,
		struct v4l2_queryctrl *qc)
{
	int i;
	for (i = 0; i < ARRAY_SIZE(radio_qctrl); i++)  	// Find which control the user is asking for
		if (qc->id && qc->id == radio_qctrl[i].id){
			memcpy(qc, &(radio_qctrl[i]), sizeof(*qc)); // If it's found, copy to the user struct and return 0
			return 0;
		}
	return -EINVAL; //control no available
}


/*  Does any radio driver support this? Neither we do.
		case VIDIOC_QUERYMENU:		// * Reports the available controls to applications (Volume, mute, etc) * /
		{
			return 0;
		}
*/

/*
 * Get a control value (Volume, mute status, etc)
 * TODO: almost DONE

 */
static int vidioc_g_ctrl (struct file *file, void *priv,
		struct v4l2_control *ctrl)
{
	struct video_device *dev = video_devdata(file);
	struct t_radio_device * radio_device = dev->priv;

	switch(ctrl->id){
	case V4L2_CID_AUDIO_VOLUME:
		ctrl->value = stlc2590_get_volume(radio_device);
		return 0;
	case V4L2_CID_AUDIO_MUTE:
		ctrl->value = stlc2590_get_mute(radio_device);
		return 0;
	}
	return -EINVAL;
}

/*
 * Set a control value (Volume, mute status, etc)
 * DONE
 */
static int vidioc_s_ctrl (struct file *file, void *priv,
		struct v4l2_control *ctrl)
{
	struct video_device *dev = video_devdata(file);
	struct t_radio_device * radio_device = dev->priv;

	switch(ctrl->id){
	case V4L2_CID_AUDIO_VOLUME:
		if(ctrl->value >= 0  && ctrl->value <=65535){
			if(stlc2590_set_volume(radio_device, ctrl->value, 1)){
				return -1; // error while sending commands
				nmdk_info("Error sending commands to set volume value");
			}
			return 0;

		} else {
			nmdk_info("Error while changing volume. User space trying to set volume to value out of bounds");
			return -ERANGE;
		}
		return 0;
	case V4L2_CID_AUDIO_MUTE:
		if(stlc2590_set_mute(radio_device, ctrl->value,1))
			return -1;
		return 0;
	}
	return -EINVAL;
}

/*
 * TODO: what's this?!?!?
 * Apparently DONE
 */
static int vidioc_g_audio (struct file *file, void *priv,
		struct v4l2_audio *a)
{
	if (a->index > 1)
		return -EINVAL;

	strcpy(a->name, "Radio");
	a->capability = V4L2_AUDCAP_STEREO;
	return 0;
}

/*
 * TODO: what's this?!?!?
 * Apparently DONE
 */
static int vidioc_s_audio (struct file *file, void *priv,
		struct v4l2_audio *a)
{
	if (a->index != 0)
		return -EINVAL;
	return 0;
}

/*
 * Get the current input. According to V4L2 specs, for a radio device this value is always 0
 * DONE
 */
static int vidioc_g_input (struct file *filp, void *priv, unsigned int *i)
{
	*i = 0;
	return 0;
}

/*
 * Set the current input. Accoding to V4L2 specs, it's not possible to set a value other than 0
 * DONE
 */
static int vidioc_s_input (struct file *filp, void *priv, unsigned int i)
{
	if (i != 0)
		return -EINVAL;
	return 0;
}


static ssize_t stlc2590_read(struct file *file, char __user *data, size_t count, loff_t *ppos)
{
	if(copy_to_user(data,rdsbuf,12))
		return -EFAULT;
	return 12;
}



//set default values

static struct t_radio_device radio_device;

static const struct file_operations stlc2590_fops = {
	.owner 		= THIS_MODULE,
	.open		= video_exclusive_open,
	.release 	= video_exclusive_release,
	.ioctl		= video_ioctl2,
	.compat_ioctl = v4l_compat_ioctl32,
	.llseek		= no_llseek,
	.read 		= stlc2590_read,
};

static struct video_device stlc2590_dev = {
	.owner 			= THIS_MODULE,
	.name  			= "FM Radio",
	.type  			= VID_TYPE_TUNER,
	.hardware 		= 0,
	.fops 			= &stlc2590_fops,
	.vidioc_querycap 	= vidioc_querycap,
	.vidioc_g_tuner		= vidioc_g_tuner,
	.vidioc_s_tuner		= vidioc_s_tuner,
	.vidioc_g_frequency	= vidioc_g_frequency,
	.vidioc_s_frequency	= vidioc_s_frequency,
	.vidioc_queryctrl	= vidioc_queryctrl,
	.vidioc_g_ctrl		= vidioc_g_ctrl,
	.vidioc_s_ctrl		= vidioc_s_ctrl,
	.vidioc_g_audio		= vidioc_g_audio,
	.vidioc_s_audio		= vidioc_s_audio,
	.vidioc_g_input		= vidioc_g_input,
	.vidioc_s_input		= vidioc_s_input,
};


static ssize_t show_stereo(struct class_device *cd, char *buf)
{
	struct video_device *vdev = to_video_device(cd);
	struct t_radio_device *dev = vdev->priv;

	return sprintf(buf, "%s\n",stlc2590_get_stereoindicator(dev)?"stereo":"mono");
}
static CLASS_DEVICE_ATTR(stereo, S_IRUGO, show_stereo, NULL);

static ssize_t show_rdstest(struct class_device *cd, char *buf)
{
//	struct video_device *vdev = to_video_device(cd);
//	struct t_radio_device *dev = vdev->priv;

	return sprintf(buf, "%d\n", rdstest);
}
static CLASS_DEVICE_ATTR(rdstest, S_IRUGO, show_rdstest, NULL);


static int stlc2590_create_sysfs(struct video_device * vdev)
{
	int rc;
	rc = video_device_create_file(vdev, &class_device_attr_stereo);
	if(rc)
		goto err_stereo;

	rc = video_device_create_file(vdev, &class_device_attr_rdstest);
	if(rc)
		goto err_rdstest;


	return 0;
err_rdstest:
	video_device_remove_file(vdev, &class_device_attr_rdstest);
err_stereo:
	video_device_remove_file(vdev, &class_device_attr_stereo);
	return rc;
}

static void stlc2590_rds_callback(void * parameter)
{
	struct t_radio_device *dev = (struct t_radio_device *) parameter;
	rdstest +=1;


	stlc2590_refresh_read(dev);


	//RDSA
	rdsbuf[0] = dev->regs[STLC2590_RDSA_LOW];
	rdsbuf[1] = dev->regs[STLC2590_RDSA_UP];
	rdsbuf[2] = 0;  // error

	//RDSB
	rdsbuf[3] = dev->regs[STLC2590_RDSB_LOW];
	rdsbuf[4] = dev->regs[STLC2590_RDSB_UP];
	rdsbuf[5] = 1;  // error

	//RDSC
	rdsbuf[6] = dev->regs[STLC2590_RDSC_LOW];
	rdsbuf[7] = dev->regs[STLC2590_RDSC_UP];
	rdsbuf[8] = 2;  // error

	//RDSD
	rdsbuf[9] = dev->regs[STLC2590_RDSD_LOW];
	rdsbuf[10] = dev->regs[STLC2590_RDSD_UP];
	rdsbuf[11] = 3;  // error

#if 0
	if ((((dev->regs[STLC2590_STATUSRSSI_UP] >> 1) & 0x03)  != 0x3) 
  		&& ((dev->regs[STLC2590_READCHAN_UP] >> 2)   != 0x3F) ) { // 0B


		if((dev->regs[STLC2590_RDSB_UP] & 0x0F ) || (dev->regs[STLC2590_RDSB_UP] & 0x07 ) ) {

			pname[((dev->regs[STLC2590_RDSB_UP] & 0x03) <<1)] = dev->regs[STLC2590_RDSD_UP];
			pname[((dev->regs[STLC2590_RDSB_UP] & 0x03) <<1) +1] = dev->regs[STLC2590_RDSD_LOW];

			nmdk_error ("pname: %s", pname);

		}



	nmdk_error ("rds groups: 0x%04x 0x%04x 0x%04x 0x%04x",
					(dev->regs[STLC2590_RDSA_UP] << 8) | dev->regs[STLC2590_RDSA_LOW],
				 	(dev->regs[STLC2590_RDSB_UP] << 8) | dev->regs[STLC2590_RDSB_LOW],
					(dev->regs[STLC2590_RDSC_UP] << 8) | dev->regs[STLC2590_RDSC_LOW],
					(dev->regs[STLC2590_RDSD_UP] << 8) | dev->regs[STLC2590_RDSD_LOW]);

	}
	
#endif
		

	STMPE2401_Acknowledge(0, 0x100, 0x2000);
}


static int stlc2590_init_irq (void (*callback)(void* parameter), void * p)
{
	t_STMPE2401_error err;
	// Configure exp gpio for rds
	
	
	err = STMPE2401_Install_Callback(0, STMPE2401_GPIO_IRQ(13), callback,(void*)p);

	err |= STMPE2401_SetGpioEdgeDetect(0, EGPIO_PIN_13,STMPE2401_FALL_EDGE);
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't set rds edge status");

	err |= STMPE2401_ClearGpioEdgeStatus(0,(0x1<<EGPIO_PIN_13));
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't clear rds edge status");

	err = STMPE2401_InterruptSourceAbilitation(0, STMPE2401_GPIO_IRQ(13), STMPE2401_ENABLE_INTERRUPT);
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't setup callback for rds");

#if 0	
	err = STMPE2401_SetGpioAltFunction(0,EGPIO_PIN_13,STMPE2401_PRIMARY_FUNCTION); 
	err |= STMPE2401_SetGpioDir(0,EGPIO_PIN_13,STMPE2401_GPIO_IN);
	err |= STMPE2401_SetGpioPull(0,EGPIO_PIN_13,STMPE2401_PULL_UP);

	err |= STMPE2401_Install_Callback(0, STMPE2401_GPIO_IRQ(13), callback,(void*)p);
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't setup callback for rds");

	err |= STMPE2401_SetGpioEdgeDetect(0, EGPIO_PIN_13,STMPE2401_FALL_EDGE);
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't set rds edge status");

	err |= STMPE2401_ClearGpioEdgeStatus(0,(0x1<<EGPIO_PIN_13));
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't clear rds edge status");


	err = STMPE2401_InterruptSourceAbilitation(0, STMPE2401_GPIO_IRQ(13), STMPE2401_ENABLE_INTERRUPT);
	if(err != STMPE2401_OK)
		nmdk_error ("Couldn't setup callback for rds");


	err = STMPE2401_InterruptAbilitation(0, STMPE2401_ENABLE_INTERRUPT);

#endif
	return (err != STMPE2401_OK);
	
}

static int stlc2590_exit_irq(void)
{
	 t_STMPE2401_error err;
         
	 err = STMPE2401_InterruptSourceAbilitation(0, STMPE2401_GPIO_IRQ(13), STMPE2401_DISABLE_INTERRUPT );
         err |= STMPE2401_SetGpioEdgeDetect(0, EGPIO_PIN_13, STMPE2401_NO_EDGE);
         return (err != STMPE2401_OK);
}

static int __init stlc2590_init (void)
{

	uint8 * regs = radio_device.regs;
	
	memset(radio_device.regs,0, sizeof(radio_device.regs[32]));
	regs[STLC2590_POWERCFG_LOW] = 0xCE; regs[4] = 0xC3; regs[8] = 0x08; regs[9] = 0x14; regs[11]= 0x1F;


	nmdk_info("Registering radio device driver");
	stlc2590_dev.priv = &radio_device;
	if(video_register_device(&stlc2590_dev, VFL_TYPE_RADIO, radio_nr)){
		nmdk_error("Impossible to register device");
		return -EINVAL;
	}
	nmdk_info("Radio device registered");

	nmdk_info("Initializing device");	

	//Set PIN 15 on Port Expander to OUT, so we can control the reset pin on 
	STMPE2401_SetGpioDir(0,EGPIO_PIN_15,STMPE2401_GPIO_OUT);
	
	//RESET PIN - Put it on LOW and after a while on HIGHT
	STMPE2401_SetGpioVal(0,EGPIO_PIN_15,0);
	msleep(100);
	STMPE2401_SetGpioVal(0,EGPIO_PIN_15,1);
	msleep(100);

	// Configure exp gpio for FM stereo
	STMPE2401_SetGpioDir(0,EGPIO_PIN_14,STMPE2401_GPIO_IN);
	

	//audiocodec
	init_audio_codec();


	stlc2590_set_mute(&radio_device, STLC2590_DEFAULT_MUTE, 1); 		//MUTE device

	//Set Enable bit
	regs[STLC2590_TEST1_UP] = 0x00;
	regs[STLC2590_POWERCFG_LOW] = 0x01;
	stlc2590_refresh_write(&radio_device);

	//wait for powerup time - refer to documentation TABLE 30, 110ms   (+ sec. threshold... =200ms)
	msleep(200);
	nmdk_info("Powered up");

	stlc2590_refresh_read(&radio_device);

	nmdk_info("Part number: 0x%x, Manufacturer ID: 0x%x",
			regs[STLC2590_DEVICEID_UP] >> 4,
			(regs[STLC2590_DEVICEID_UP]&0x0F)|regs[STLC2590_DEVICEID_LOW]
			);
	nmdk_info("Chip version: 0x%x, Device: 0x%x, Firmware version: 0x%x",
			regs[STLC2590_CHIPID_UP] >> 2,
			(regs[STLC2590_CHIPID_UP] >> 1) & 0x01,
			(regs[STLC2590_CHIPID_UP] & 0x03) | regs[STLC2590_CHIPID_LOW]
			);



	stlc2590_set_volume(&radio_device, STLC2590_DEFAULT_VOLUME, 0); 	//Set volume to max
	stlc2590_set_channel_spacing(&radio_device, channel_spacing, 0);  	//Channel spacing  == 100 Khz (Europe)
	stlc2590_set_band_limits(&radio_device, band_limits, 0);		//Set band limits - Europe (87.5kHz - 104kHz)
	stlc2590_set_rdsenable(&radio_device, rds_enable, 0);			//Disable RDS
	stlc2590_set_stereoindicator(&radio_device, 1, 0);			//Enable stereo indicator
	stlc2590_set_stereoblend(&radio_device, stereo_blend, 0);

	if (stlc2590_refresh_write(&radio_device))
		goto ERROR_SENDING_CTRL;

	stlc2590_tune_frequency(&radio_device, 10250);
	printregs(regs);

	//Create sysfs files
	if(stlc2590_create_sysfs(&stlc2590_dev))
		nmdk_error("Unable to create sysfs");


	//Set interrupts
	if(stlc2590_init_irq(&stlc2590_rds_callback,&radio_device))
		nmdk_error("Unable to set rds interrupt");

	return 0;
ERROR_SENDING_CTRL:
	nmdk_error("Error sending initialization controls");
	video_unregister_device(&stlc2590_dev);
	return -EINVAL;
}

MODULE_AUTHOR("Lucas De Marchi, Cristina Chang, Patrick Bellasi");
MODULE_DESCRIPTION("STLC2590 radio driver for NHK-1.5\'s board");
MODULE_LICENSE("Dual BSD/GPL");


static void __exit stlc2590_exit (void)
{

	if(stlc2590_exit_irq())
		nmdk_error("Unable to remove irq");

	//Disable external amplifiers
	STMPE2401_SetGpioDir(0,EGPIO_PIN_12,STMPE2401_GPIO_OUT);
	STMPE2401_SetGpioVal(0,EGPIO_PIN_12,0);
	
	// Set AHIZEN bit - Silent shutdown
	radio_device.regs[STLC2590_TEST1_UP] = 0x40;
	stlc2590_refresh_write(&radio_device);
	msleep(20);

	//power down device
	radio_device.regs[STLC2590_POWERCFG_UP] = 0x40;
	stlc2590_refresh_write(&radio_device);
	//msleep(120);

	//disable device
	STMPE2401_SetGpioVal(0,EGPIO_PIN_15,0);

	//unregister it from V4L2
	video_unregister_device(&stlc2590_dev);

	printk(KERN_INFO "Goodbye, cruel world\n");
	
}

module_init(stlc2590_init);
module_exit(stlc2590_exit);

